;       AX.25 relay routine                        N0QBH   2000
;       Version 2.2
;       10 Mhz XTAL
;       Ramtron serial 2k RAM installed
;       Inserts repeater call over alias
;       MAX storage     255 bytes
;       SW1 toggles the data filter L=WIDE R=ALL
;       Wake up wink LED
;       Rx uses RBI Flag
;       5/15/00
;------------------------------------------------------------
	LIST	P=PIC16F84
        	include	"P16F84.INC"
        __CONFIG _HS_OSC & _CP_OFF & _PWRTE_ON & _WDT_OFF
;-------------------------------------------------------------
;               Port Set-up
;           0   1   2   3   4   5   6   7       bit
; port  A   Out Out Out Out In  X   X   X
;       B   Out Out Out In  In  Out Out Out
;------------------------------------------------------------
	ORG	0			; start code
	goto	main
;------------------------------------------------------------
my_aliases
        addwf   PCL,F
        dt      "WIDE  ",0              ;must be 6 char long
;------------------------------------------------------------
my_call_sign
        addwf   PCL,F
        dt      "N0QBH ",0              ;must be 6 char long
;------------------------------------------------------------
        cblock  0x0c
                mirror_reg              ;buffer
                aux_count               ;flag counter
                byte_count              ;delay counter
                store_count             ;bytes stored cnt
                timer_count             ;timer counter
                bit_stuff_count         ;counter
                bit_count               ;counter
                send_reg                ;packet data
                flag_count              ;flag counter
                FLAGS                   ;various flags
                byte_in                 ;data byte rx'd
                AX25_crc_hi             ;upper byte CRC word
                AX25_crc_lo             ;lower byte CRC word
                next_to_last_byte       ;2nd from last byte
                last_byte               ;last byte rxed
                next_freq               ; 16 toggles 0/1
                ssid_buffer             ;SSID storage
                RAMAddressHigh          ;Upper 3 bits
                RAMAddressLow           ;lower 8 bits
                RAMSave                 ;buffer
                RAMData                 ;data byte
                RAMCount                ;ram counter
        endc
;------------------------------------------------------------
#define RAM_DATA_IN             PORTA,0         ;RAM SO
#define RAM_DATA_OUT            PORTA,0         ;RAM SI
#define PTT_OUT                 PORTA,2         ;radio PTT
#define OPTION_TOG              PORTA,4         ;SW1
#define VISUAL_ONE              PORTB,0         ;LED
#define MODE1_OUT               PORTB,1         ;RX/TX mode
#define MODE0_OUT               PORTB,2         ;RX/TX/lo pwr
#define DCD_IN                  PORTB,3         ;dcd input
#define PKT_DATA_OUT            PORTB,5         ;tx data
#define RAM_CLOCK               PORTB,6         ;RAM SCK
#define RAM_CHIP_SELECT         PORTB,7         ;RAM CS
#define DATA_FLAG               FLAGS,0         ;beg/end flag
#define RAM_WRITE_FLAG          FLAGS,1         ;RAM setup?
#define TOGGLE_FLAG             FLAGS,2         ;SW1 input
;------------------------------------------------------------
#define RAM_OP_WREN             B'00000110'
#define	RAM_OP_READ		B'00000011'
#define	RAM_OP_WRITE  		B'00000010'
;------------------------------------------------------------
main
        bsf     STATUS,RP0              ;bank 1
        movlw   b'11110000'             ;A:4 input
        movwf   TRISA                   ;set up port A I/O
        movlw   b'00011000'             ;B:3,:4,inputs
        movwf   TRISB                   ;set up port B I/O
        movlw   b'11010011'
        movwf   OPTION_REG              ;TMR0 /16
        bcf     STATUS,RP0              ;bank 0
        clrf    INTCON                  ;clear INT reg
        bcf     PTT_OUT                 ;unkey radio
        nop
        bsf     MODE0_OUT               ;MX614 setup
        nop
        bsf     MODE1_OUT               ;lo pwr mode
        movlw   0x0a                    ;10 loops
        call    adjustable_delay        ;2 mS delay
        bcf     MODE0_OUT               ;MX614 RX mode
        movlw   0x64                    ;100 loops
        call    adjustable_delay        ;20 mS delay
        call    wake_up_wink            ;flash LED
        goto    house_clean
;------------------------------------------------------------
back_from_tx
        call    mx614_receive
house_clean                             ;new packet
        call    RAM_init                ;init RAM
        clrf    aux_count               ;reset misc counter
        clrf    byte_in                 ;clear data byte
        clrf    bit_stuff_count         ;reset b_s_count
        clrf    store_count             ;reset count out
        clrf    byte_count              ;reset byte count
        clrf    FLAGS                   ;reset flags
        movlw   0xff
        movwf   AX25_crc_hi             ;reset CRC calc
        movwf   AX25_crc_lo
        movlw   0x7e
        movwf   TMR0
        movfw   PORTB                   ;reset bit change
        clrf    INTCON                  ;set up TMR0, RBI
        bcf     VISUAL_ONE              ;clear flags lite
        btfsc   OPTION_TOG
        bsf     TOGGLE_FLAG             ;use alias one
find_start_flag                         ;use alias two
        clrc
        call    recv_one_bit            ;check for toggle
        skpnz
        setc                            ;no toggle (1)
        rrf     byte_in,F               ;toggle  (0)
        movfw   byte_in                 ;byte_in to W
        xorlw   0x7e                    ;is it a flag ?
        skpz
        goto    find_start_flag         ;no
start_flag_loop                         ;yes, look again
        call    recv_one_bit
        skpnz                           ;see a toggle ?
        goto    find_start_flag         ;no, go back
        movlw   0x06                    ;yes, must see 6 1's
        movwf   store_count
long_ones
        call    recv_one_bit            ;look for 6
        skpz                            ;1's in arow
        goto    find_start_flag         ;saw a 0
        decfsz  store_count,F           ;saw a 1
        goto    long_ones
        call    recv_one_bit            ;look for end 0
        skpnz
        goto    find_start_flag         ;not flag, reset 
;------------------------------------------------------------
;
get_next_byte
        clrf    byte_in                 ;clear input byte
        movlw   0x08
        movwf   bit_count               ;get 8 bits
get_next_bit
        call    recv_one_bit
        skpz
        goto    zero_in                 ; 0 rx'd
        incf    bit_stuff_count,F       ;inc counter
        setc                            ; 1 rx'd
roll_em_in
        rrf     byte_in,F               ;bit stored
        movfw   bit_stuff_count
        xorlw   0x05                    ;5 1's in a row ?
        skpnz
        goto    check_stuff             ;yes
count_em_out                            ;no
        decfsz  bit_count,F             ;8 bits ?
        goto    get_next_bit            ;no
        goto    CRC_save_store          ;yes
zero_in
        clrf    bit_stuff_count         ;0 seen, clear bsc
        clrc
        goto    roll_em_in
;------------------------------------------------------------
; Bit stuff decode routine
;
check_stuff
        call    recv_one_bit            ;wait 1 bit
        skpnz                           ;if data=0, skip next
        goto    sixth_one               ;unchanged, flag ?
        clrf    bit_stuff_count         ;tog seen, clear count
        goto    count_em_out            ;don't count this bit
sixth_one
        call    recv_one_bit            ;get 6th bit
        skpnz                           ;flag ?
        goto    house_clean             ;no, 7th 1, reset
        movfw   bit_count
        xorlw   0x03
        skpz                            ;in the right place ?
        goto    house_clean             ;no
        clrf    bit_stuff_count         ;yes, flag, clear bstc
        btfss   DATA_FLAG               ;header/tail flag ?
        goto    get_next_byte           ;header flag
        movfw   byte_count              ;tail flag
        sublw   0x14                    ;how long is packet ?
        skpnc
        goto    house_clean             ; < 21 bytes rx'd
        goto    packet_end              ;it's long enough
;
CRC_save_store
        btfss   DATA_FLAG               ;test data flag
        bsf     DATA_FLAG               ;set data flag
        incf    byte_count,F            ;count bytes
        movfw   byte_count
        xorlw   0xff                    ;how long ?
        skpnz
        goto    house_clean             ; 255 bytes, reset
        movfw   byte_count              ;how long again
        sublw   0x02
        skpc                            ; >2 ?
        call    AX25_CRC_calc           ;
        movfw   last_byte               ;last byte to W
        movwf   next_to_last_byte       ;to 2nd to last
        movfw   byte_in                 ;get new data byte
        movwf   last_byte               ;store in last byte
        goto    get_next_byte           ;get next byte
;
;       Invert CRC data and compare to value sent
;
packet_end
        call    RAMStop                 ;stop Ram write
        movlw   0xff
        xorwf   AX25_crc_lo,F
        xorwf   AX25_crc_hi,F           ;invert  CRC data
        movfw   next_to_last_byte       ;check lobyte match
        xorwf   AX25_crc_lo,W
        skpz
        goto    house_clean             ;mismatch, reset
        movfw   last_byte
        xorwf   AX25_crc_hi,W           ;check hibyte match
        skpz
        goto    house_clean             ;mismatch, reset
        btfss   TOGGLE_FLAG
        goto    transmit_set_up         ;no filter
;------------------------------------------------------------
valid_digi_check
        movlw   0x0d
        movwf   RAMAddressLow
        call    RAMReadByte
        movwf   ssid_buffer
        btfsc   ssid_buffer,0           ;any digis ?
        goto    house_clean             ;no
        movlw   0x14                    ;yes
        movwf   RAMAddressLow           ;1st digi ssid
F_flag_loop
        call    RAMReadByte
        andlw   0x81
        movwf   ssid_buffer             ;W to ssid buffer
        xorlw   0x81
        skpnz
        goto    house_clean             ;last addr, F bit set
        btfss   ssid_buffer,7           ;F bit set ?
        goto    locate_address          ;no
        movlw   0x07                    ;yes, look at next
        addwf   RAMAddressLow,W
        sublw   0x30
        skpc                            ; > 7 loops ?
        goto    house_clean             ;yes
        movlw   0x07                    ;no
        addwf   RAMAddressLow,F
        goto    F_flag_loop
locate_address
        movlw   0x06                    ;no
        subwf   RAMAddressLow,F         ;go back 6 bytes
        clrf    aux_count
look_for_match                          ;compare to alias
        movfw   aux_count               ;aux_count to W
        call    my_aliases              ;look up table
        movwf   mirror_reg              ;store table data
        xorlw   0x00                    ;see end ?
        skpnz
        goto    relay_packet            ;yes
        clrc                            ;no
        rlf     mirror_reg,F
        call    RAMReadByte             ;read 1 byte
        xorwf   mirror_reg,W            ;same ?
        skpz
        goto    house_clean             ;no match
        incf    RAMAddressLow,F         ;match
        incf    aux_count,F
        goto    look_for_match
;------------------------------------------------------------
;               Overwrites alias with my_call
relay_packet
        movlw   0x06
        subwf   RAMAddressLow,F         ;start of call
        clrf    aux_count
my_call_loop
        movfw   aux_count
        call    my_call_sign            ;get table data
        movwf   mirror_reg              ;store data
        xorlw   0x00
        skpnz
        goto    set_f_bit               ;done
        clrc
        rlf     mirror_reg,W            ;left shift 1
        call    RAMWriteByte
        incf    RAMAddressLow,F
        incf    aux_count,F
        goto    my_call_loop
set_f_bit
        call    RAMReadByte             ;get SSID
        iorlw   0x80                    ;set F bit
        andlw   0x81                    ;make it 0
        call    RAMWriteByte
transmit_set_up
        clrf    send_reg
        clrf    bit_count
        clrf    bit_stuff_count
        clrf    FLAGS
        clrf    next_freq
;
check_dcd                               ;send packet
        call    twoh_uS_delay           ;wait .2 mS
        btfsc   DCD_IN                  ;check DCD pin
        goto    check_dcd               ;for clear channel
        movlw   0xff
        movwf   AX25_crc_hi             ;reset CRC calc
        movwf   AX25_crc_lo
        call    mx614_transmit          ;set up for tx
        bcf     PKT_DATA_OUT            ;start tone = 0
        bsf     PTT_OUT                 ;key radio
        movlw   0x7e
        movwf   TMR0
        bcf     INTCON,T0IF             ;set up timer
        movlw   0x28
        movwf   flag_count              ;40 flag delay
        call    tx_flags                ;send flags for tx
;
        clrf    RAMAddressLow           ;start of data loc
        call    RAMReadStart            ;start read at loc
        movlw   0x02
        subwf   byte_count,F            ;byte_count - 2
transmit_data_loop
        call    RAMRead
        movwf   send_reg                ;W to send_reg
        call    tx_one_byte
        decfsz  byte_count,F            ;count down to 0
        goto    transmit_data_loop
wrap_it_up
        call    RAMStop
        movlw   0xff
        xorwf   AX25_crc_hi,W           ;invert both
        movwf   next_to_last_byte       ;CRC bytes
        movlw   0xff
        xorwf   AX25_crc_lo,W
        movwf   send_reg
        call    tx_one_byte             ;send CRC lo byte
        movfw   next_to_last_byte
        movwf   send_reg
        call    tx_one_byte             ;send CRC hi byte
        movlw   0x02
        movwf   flag_count
        call    tx_flags                ;send 2 end flags
        bcf     PTT_OUT                 ;unkey radio
        goto    back_from_tx            ;done, start over
;
;------------------------------------------------------------
;       Subroutines area
;------------------------------------------------------------
RAM_init
        incf    RAMAddressHigh,F        ;inc high addr byte
        movfw   RAMAddressHigh
        sublw   0x07
        skpc                            ; > 7 ?
        clrf    RAMAddressHigh          ;yes, clear
        clrf    RAMAddressLow           ;no
        clrf    RAMData
        bsf     RAM_CHIP_SELECT
        bcf     RAM_CLOCK
        return
;------------------------------------------------------------
RAMWrite
        bsf     STATUS,RP0              ;bank 1
        bcf     RAM_DATA_OUT            ;make A:0 output
        bcf     STATUS,RP0              ;bank 0
        movwf   RAMData                 ;W is data
        movlw   D'8'
        movwf   RAMCount                ;8 bits
RAMSendLoop
	rlf	RAMData, F
	bcf	RAM_DATA_OUT
	skpnc
        bsf     RAM_DATA_OUT
	bsf	RAM_CLOCK
	bcf	RAM_CLOCK
	decfsz	RAMCount, F
        goto    RAMSendLoop
	return
;------------------------------------------------------------
RAMRead
        bsf     STATUS,RP0              ;bank 1
        bsf     RAM_DATA_IN             ;make A:0 input
        bcf     STATUS,RP0              ;bank 0
	movlw	D'8'
	movwf	RAMCount
RAMGetLoop
	bsf	RAM_CLOCK
	clrc
	btfsc	RAM_DATA_IN
        setc
	rlf	RAMData, F
	bcf	RAM_CLOCK
	decfsz	RAMCount, F
        goto    RAMGetLoop
	movfw	RAMData
        return
;------------------------------------------------------------
RAMWriteEnable
	bcf	RAM_CHIP_SELECT
	movlw	RAM_OP_WREN
	call	RAMWrite
RAMStop
	bsf	RAM_CHIP_SELECT
	return
;------------------------------------------------------------
RAMWriteStart
	call	RAMWriteEnable
	bcf	RAM_CHIP_SELECT
	swapf	RAMAddressHigh, W
	movwf	RAMData
	rrf	RAMData, W
	andlw	B'00111000'
	iorlw	RAM_OP_WRITE
	call	RAMWrite
	movfw	RAMAddressLow
	call	RAMWrite
	return
;
RAMWriteByte
	movwf	RAMSave
	call	RAMWriteStart
	movfw	RAMSave
	call	RAMWrite
	goto	RAMStop
;------------------------------------------------------------
RAMReadStart
	bcf	RAM_CHIP_SELECT
	swapf	RAMAddressHigh, W
	movwf	RAMData
	rrf	RAMData, W
	andlw	B'00111000'
	iorlw	RAM_OP_READ
	call	RAMWrite
	movfw	RAMAddressLow
	call	RAMWrite
	return
;
RAMReadByte
	call	RAMReadStart
	call	RAMRead
        goto    RAMStop
;------------------------------------------------------------
insert_id
        clrf    aux_count
id_loop
        movfw   aux_count
        call    my_call_sign            ;get table data
        movwf   mirror_reg              ;store data
        xorlw   0x00
        skpnz
        return
        movfw   mirror_reg
        movwf   send_reg
        call    tx_one_byte
        goto    id_loop
;------------------------------------------------------------
mx614_receive
        bcf     MODE0_OUT
        nop
        bsf     MODE1_OUT
        movlw   0x0a                    ;2 mS delay
        call    adjustable_delay        ;rx mode set
        return
;------------------------------------------------------------
mx614_transmit
        bcf     MODE1_OUT
        nop
        bsf     MODE0_OUT               ;MX614 tx
        movlw   0x02
        call    adjustable_delay        ;wait .4 mS
        return
;------------------------------------------------------------
;       Stores all packet data as received
;
AX25_CRC_calc
        btfsc   RAM_WRITE_FLAG          ;are we setup ?
        goto    FRAM_store              ;yes
        clrf    RAMAddressLow           ;no, start
        call    RAMWriteStart           ;loc to RAM
        bsf     RAM_WRITE_FLAG          ;setup done
        bsf     VISUAL_ONE              ;LED on
;       RAM data store routine
;
FRAM_store
        movfw   next_to_last_byte       ;get byte
        call    RAMWrite                ;store in RAM
;       CRC calculation
crc_rotor
        movlw   0x08
        movwf   bit_count               ; 8 loops
crc_head
        rrf     next_to_last_byte,F     ; get bit for CRC
        call    crc_core                ;do CRC calc
        decfsz  bit_count,F             ;8 counts
        goto    crc_head
        return
;------------------------------------------------------------
crc_core
        movlw   1
        skpnc
        xorwf   AX25_crc_lo, F          ; crc calculation
        clrc
        rrf     AX25_crc_hi, F
        rrf     AX25_crc_lo, F
        skpc
        return
        movlw   H'08'
        xorwf   AX25_crc_lo, F
        movlw   H'84'
        xorwf   AX25_crc_hi, F
        return
;------------------------------------------------------------
wake_up_wink
        movlw   0x10
        movwf   flag_count
        bsf     VISUAL_ONE              ;LED on
wake_up_wink_loop
        movlw   0xff
        call    adjustable_delay        ;wait
        bcf     VISUAL_ONE
        movlw   0xff
        call    adjustable_delay
        bsf     VISUAL_ONE
        decfsz  flag_count,F
        goto    wake_up_wink_loop
        bcf     VISUAL_ONE
        return
;------------------------------------------------------------
;       Packet bit receive routine
;       Returns with 0 or 1 in Zero Flag
recv_one_bit                            ;delay 833 uS
        btfsc   INTCON,RBIF
        goto    bit_edge                ;saw tone toggle
        btfss   INTCON,T0IF
        goto    recv_one_bit            ;not timed out
        movlw   0x7e                    ;timed out
        movwf   TMR0                    ;reset timer
        bcf     INTCON,T0IF             ;1 byte long
        clrw                            ;Z flag set
        return
bit_edge
        movlw   0x3d                    ;load 1 1/2 bits
        movwf   TMR0
        movfw   PORTB                   ;reset port change
        bcf     INTCON,RBIF             ;clear RBIF
        bcf     INTCON,T0IF
        movfw   TMR0                    ;Z flag clear
        return
;------------------------------------------------------------
;               Assumes byte is in send_reg
tx_one_byte
        movlw   0x08
        movwf   bit_count               ;8 bits
rotor_out
        rrf     send_reg,W              ;calc CRC
        call    crc_core
        rrf     send_reg,F              ;get bit to send
        skpc
        goto    zero_out
        incf    bit_stuff_count,F       ; b_s_c +1
        call    tx_one_bit              ;send a 1
        movfw   bit_stuff_count         ;check b_s_c
        xorlw   0x05
        skpnz                           ;b_s_f = 5 ?
        goto    zero_out                ;yes
count_the_bits
        decfsz  bit_count,F
        goto    rotor_out
        return
zero_out
        clrf    bit_stuff_count         ;reset b_s_f
        incf    next_freq,F
        call    tx_one_bit              ;send a 0
        goto    count_the_bits          ;
;
;------------------------------------------------------------
;               Sends 01111110 flag_count times
tx_flags
        movlw   0x06
        movwf   bit_count               ;8 bits
        incf    next_freq,F
        call    tx_one_bit              ;send 0
flag_twist
        call    tx_one_bit              ;send 6 1's
        decfsz  bit_count,F
        goto    flag_twist
        incf    next_freq,F
        call    tx_one_bit              ;send 0
        decfsz  flag_count,F
        goto    tx_flags
        return                          ;end of flags
;
;------------------------------------------------------------
;
tx_one_bit                              ;832 uS delay
        btfss   INTCON,T0IF
        goto    tx_one_bit
        movfw   next_freq               ;next_freq to W
        andlw   0x01                    ;mask upper 7 bits
        skpz                            ;0 or 1 ?
        goto    lo_tone                 ;1, goto lo_tone
        bcf     PKT_DATA_OUT            ;0, send 2200 Hz
        goto    reset_the_timer
lo_tone
        bsf     PKT_DATA_OUT            ;send 1200 Hz
reset_the_timer                         ;one, no change
        movlw   0x7e                    ;832uS
        movwf   TMR0
        bcf     INTCON,T0IF             ;clr timeout flag
        return
;------------------------------------------------------------
twoh_uS_delay                           ;200 uS delay
        movlw   0xa6                    ;166 loops
        movwf   timer_count
twoh_uS_loop
        decfsz  timer_count,F
        goto    twoh_uS_loop
        return
;------------------------------------------------------------
adjustable_delay                        ;W contains length
        movwf   aux_count
the_mS_loop
        call    twoh_uS_delay
        decfsz  aux_count,F
        goto    the_mS_loop
        return
;------------------------------------------------------------
        END



